import { readFileSync } from "node:fs";
import { basename, dirname, extname, join } from "node:path";
import { once } from "lodash";
import hash from "@emotion/hash";
import type babel from "@babel/core";
import type * as vite from "vite";
import type * as esbuild from "esbuild";
import { type NodePath, type types as t, transformSync } from "@babel/core";
import tailwind, { type Config } from "tailwindcss";
import postcss from "postcss";

const PLUGIN_NAME = "tailwind";
const ESBUILD_NAMESPACE = "babel-tailwind";
const ROLLUP_PREFIX = "\0tailwind:";

const { name } = [require][0](
  process.env.BABEL_TAILWIND_BUILD ? "./package.json" : "../package.json"
);

const definePlugin =
  <T>(fn: (runtime: typeof babel) => babel.Visitor<babel.PluginPass & T>) =>
  (runtime: typeof babel) => {
    const plugin: babel.PluginObj<babel.PluginPass & T> = {
      visitor: fn(runtime),
    };
    return plugin as babel.PluginObj;
  };

const extractJSXContainer = (attr: NonNullable<t.JSXAttribute["value"]>): t.Expression =>
  attr.type === "JSXExpressionContainer" ? (attr.expression as t.Expression) : attr;

function matchPath(
  nodePath: NodePath<t.Node | null | undefined>,
  fns: (dig: (nodePath: NodePath) => void) => babel.Visitor
) {
  if (!nodePath.node) return;
  const fn = fns(path => matchPath(path, fns))[nodePath.node.type] as any;
  fn?.(nodePath);
}

/**
 * Tagged template macro function for Tailwind classes
 * @example "tw" => tw`p-2 text-center`
 */
export type TaggedTailwindFunction = (strings: TemplateStringsArray) => string;

const tailwindDirectives = ["components", "utilities", "variants"] as const;

export interface TailwindPluginOptions {
  /**
   * Tailwind CSS configuration
   */
  tailwindConfig: Omit<Config, "content">;

  /**
   * Directives to prefix to all Tailwind stylesheets
   */
  directives?: (typeof tailwindDirectives)[number][] | "all";

  /**
   * Additional PostCSS plugins (optional)
   */
  postCSSPlugins?: postcss.AcceptedPlugin[];

  /**
   * Attribute to use for tailwind classes in JSX
   * @default "css"
   */
  jsxAttributeName?: string;

  /**
   * Tagged template macro function to use for Tailwind classes
   * @default "tw"
   * @example
   * declare const tw: TaggedTailwindFunction;
   * "tw" => tw`p-2 text-center`
   */
  taggedTemplateName?: string;

  /**
   * The prefix to use for the generated class names.
   * @default className => `tw-${hash(className)}`
   */
  getClassName?: GetClassName;

  /**
   * Preferred library for classnames
   */
  clsx: "clsx" | "classnames" | "emotion";
}

type GetClassName = (className: string) => string;

/**
 * Hashes and prefixes a string of Tailwind class names.
 * @example getClassName("p-2 text-center") // "tw-1r6fxxz"
 */
export const getClassName: GetClassName = cls => "tw-" + hash(cls);

interface BabelPluginState {
  getCx: () => t.Identifier;
  tailwindMap: Map<string, string>;
}

function babelTailwind(
  styleMap: Map<string, string>,
  {
    clsx,
    getClassName: getClass = getClassName,
    taggedTemplateName,
    jsxAttributeName = "css",
  }: TailwindPluginOptions
) {
  function getClsxImport(t: typeof babel.types, cx: t.Identifier) {
    switch (clsx) {
      case "emotion":
        return t.importDeclaration(
          [t.importSpecifier(cx, t.identifier("cx"))],
          t.stringLiteral("@emotion/css")
        );
      case "clsx":
        return t.importDeclaration(
          [t.importDefaultSpecifier(cx)],
          t.stringLiteral("clsx")
        );
      case "classnames":
        return t.importDeclaration(
          [t.importDefaultSpecifier(cx)],
          t.stringLiteral("classnames")
        );
      default:
        throw new Error("Unknown clsx library");
    }
  }

  return definePlugin<BabelPluginState>(({ types: t }) => ({
    Program: {
      enter(path, state) {
        let cx: t.Identifier;
        state.tailwindMap = new Map();
        state.getCx = () => {
          if (cx == null) {
            cx = path.scope.generateUidIdentifier("cx");
            path.node.body.unshift(getClsxImport(t, cx));
          }
          return t.cloneNode(cx);
        };
      },

      exit({ node }, { filename, tailwindMap }) {
        if (!tailwindMap.size) return;
        if (!filename) {
          throw new Error("babel: missing state.filename");
        }

        const cssName = basename(filename, extname(filename)) + ".css";
        node.body.unshift(
          t.importDeclaration([], t.stringLiteral(`tailwind:./${cssName}`))
        );

        const result = Array.from(tailwindMap)
          .map(([className, value]) => `.${className} {\n  @apply ${value}\n}`)
          .join("\n");

        styleMap.set(join(dirname(filename), cssName), result);
      },
    },

    TaggedTemplateExpression(path, { tailwindMap }) {
      if (taggedTemplateName == null) return;

      const {
        tag,
        quasi: { quasis, expressions },
      } = path.node;
      if (!t.isIdentifier(tag, { name: taggedTemplateName })) return;

      if (expressions.length) {
        throw new Error(`${taggedTemplateName}\`\` should not contain expressions`);
      }

      const value = quasis[0].value.cooked;
      if (value) {
        const trimmed = value.replace(/\s+/g, " ").trim();
        const className = getClass(trimmed);
        tailwindMap.set(className, trimmed);
        path.replaceWith(t.stringLiteral(className));
      }
    },

    JSXAttribute(path, { tailwindMap, getCx }) {
      const { name } = path.node;
      const valuePath = path.get("value");
      if (name.name !== jsxAttributeName || !valuePath.node) return;

      const parent = path.parent as t.JSXOpeningElement;
      const classNameAttribute = parent.attributes.find(
        (attr): attr is t.JSXAttribute =>
          t.isJSXAttribute(attr) && attr.name.name === "className"
      );

      matchPath(valuePath, go => ({
        StringLiteral(path) {
          const { value } = path.node;
          if (value) {
            const trimmed = value.replace(/\s+/g, " ").trim();
            const className = getClass(trimmed);
            tailwindMap.set(className, trimmed);
            path.replaceWith(t.stringLiteral(className));
          }
        },
        JSXExpressionContainer(path) {
          go(path.get("expression"));
        },
        ConditionalExpression(path) {
          go(path.get("consequent"));
          go(path.get("alternate"));
        },
        LogicalExpression(path) {
          go(path.get("right"));
        },
        CallExpression(path) {
          for (const arg of path.get("arguments")) {
            go(arg);
          }
        },
      }));

      if (classNameAttribute) {
        const attrValue = classNameAttribute.value!;

        // If both are string literals, we can merge them directly here
        if (t.isStringLiteral(attrValue) && t.isStringLiteral(valuePath.node)) {
          attrValue.value +=
            (attrValue.value.at(-1) === " " ? "" : " ") + valuePath.node.value;
        } else {
          classNameAttribute.value = t.jsxExpressionContainer(
            t.callExpression(getCx(), [
              extractJSXContainer(attrValue),
              extractJSXContainer(valuePath.node),
            ])
          );
        }
      } else {
        parent.attributes.push(
          t.jsxAttribute(t.jsxIdentifier("className"), valuePath.node)
        );
      }
      path.remove();
    },
  }));
}

/**
 * An esbuild plugin that processes files with Babel if `getPlugins` returns any plugins.
 */
export const babelPlugin = ({
  filter = /\.[jt]sx?$/,
  getPlugins,
}: {
  filter?: RegExp;
  getPlugins(file: { path: string; contents: string }): babel.PluginItem[];
}): esbuild.Plugin => ({
  name: "babel-plugin",
  setup(build) {
    build.onLoad({ filter }, ({ path }) => {
      const load = once(() => readFileSync(path, "utf-8"));
      const plugins = getPlugins({
        path,
        get contents() {
          return load();
        },
      });

      if (!plugins.length) {
        return;
      }

      const { code } = transformSync(load(), {
        parserOpts: {
          plugins: ["jsx", "typescript"],
        },
        filename: path,
        plugins,
      })!;

      return {
        contents: code!,
        loader: extname(path).slice(1) as "js" | "ts",
      };
    });
  },
});

type Compile = ReturnType<typeof createPostCSS>;

/** @internal */
export function createPostCSS({
  tailwindConfig,
  postCSSPlugins = [],
  directives,
}: Pick<TailwindPluginOptions, "tailwindConfig" | "postCSSPlugins" | "directives">) {
  const post = postcss([
    tailwind({
      ...tailwindConfig,
      content: [{ raw: "<br>", extension: "html" }],
    }),
    ...postCSSPlugins,
  ]);

  const appliedDirectives = directives === "all" ? tailwindDirectives : directives;
  const directiveTexts = appliedDirectives?.map(d => `@tailwind ${d};\n`).join("") ?? "";

  return (css: string) => post.process(directiveTexts + css, { from: undefined });
}

const esbuildPlugin = (
  styleMap: Map<string, string>,
  compile: Compile
): esbuild.Plugin => ({
  name: PLUGIN_NAME,

  setup(build) {
    build.onResolve({ filter: /^tailwind:.+\.css$/ }, ({ path, importer }) => {
      const resolved = join(dirname(importer), path.replace(/^tailwind:/, ""));
      if (styleMap.has(resolved)) {
        return {
          path: resolved,
          namespace: ESBUILD_NAMESPACE,
        };
      }
    });

    build.onResolve({ filter: RegExp(`^${name}/base$`) }, () => ({
      path: "directive:babel",
      namespace: ESBUILD_NAMESPACE,
    }));

    build.onLoad({ filter: /.*/, namespace: ESBUILD_NAMESPACE }, async ({ path }) => {
      if (path === "directive:babel") {
        return {
          contents: (await compile(`@tailwind base;`)).css!,
          loader: "css",
        };
      }

      if (!styleMap.has(path)) return;
      const result = await compile(styleMap.get(path)!);

      return {
        contents: result.css,
        loader: "css",
      };
    });
  },
});

const vitePlugin = (styleMap: Map<string, string>, compile: Compile): vite.Plugin => ({
  name: "tailwind",
  resolveId(id, importer) {
    if (id.startsWith("tailwind:")) {
      const resolved = join(dirname(importer!), id.slice("tailwind:".length));
      if (styleMap.has(resolved)) {
        return {
          id: ROLLUP_PREFIX + resolved,
          moduleSideEffects: true,
        };
      }
    }
  },
  async load(id: string) {
    if (id.startsWith(ROLLUP_PREFIX)) {
      const resolved = id.slice(ROLLUP_PREFIX.length);
      if (styleMap.has(resolved)) {
        const result = await compile(styleMap.get(resolved)!);
        return result.css;
      }
    }
  },
});

/**
 * Main entry. Returns the plugins and utilities for processing Tailwind
 * classNames in JS.
 *
 * @example
 * import { build } from "esbuild";
 * import getTailwindPlugins, { babelPlugin } from "babel-tailwind";
 *
 * const tailwind = getTailwindPlugins(options);
 * const result = await build({
 *   plugins: [
 *     babelPlugin({ getPlugins: () => [tailwind.babel] }),
 *     tailwind.esbuild,
 *   ],
 * });
 */
export function getTailwindPlugins(options: TailwindPluginOptions) {
  const styleMap = new Map<string, string>();
  const compile = createPostCSS(options);

  return {
    compile,
    babel: babelTailwind(styleMap, options),
    esbuild: esbuildPlugin(styleMap, compile),
    vite: vitePlugin(styleMap, compile),
    styleMap,
    [Symbol.dispose]() {
      styleMap.clear();
    },
  };
}
